# SCRIPT FOR LOGGING BLUE IRIS MEMORY USAGE: BOTH SERVER AND PROCESSES
#
# PURPOSE: THIS SCRIPT LOGS BLUE IRIS MEMORY USAGE PARAMETERS TO A CSV FILE.
# THESE PARAMETERS ARE OBTAINED VIA POWERSHELL INSTEAD OF QUERYING THE BLUE IRIS 
# WEBSERVER (WHICH CAN BECOME ERRATIC DURING A MEMORY LEAK).
# THE CSV FILE IS INTERROGATED BY A SCHEDULED TASK THAT IS EXECUTING POWERSHELL SCRIPT 'BI_memleak_wdog.ps1'.
#
# THIS SCRIPT CREATES A 'RAM USAGE LOG' W/ THE FOLLOWING STRUCTURE:
#
#		Timestamp,SoftwareUsedRamBytes,Uptime,RamPct,Tag,Host
#		2023/01/01_02:00:01,906797056,0:13:58:29,66%,"periodic","TaskScheduler"
#		2023/01/01_02:06:00,1680392192,0:14:04:28,71%,"periodic","TaskScheduler"
#		2023/01/01_02:12:00,2056347648,0:14:10:28,73%,"periodic","TaskScheduler"
#		2023/01/01_02:18:03,2370572288,0:14:16:31,75%,"periodic","TaskScheduler"
#
#       Column 'SoftwareUsedRamBytes' is the memory used by the Blue Iris process
#       Column 'Uptime' is the uptime for the Blue Iris process
#       Column 'RamPct' is the memory used by the system
#       Column 'Tag' is used to identify console- and simulation-created entries
#       Column 'Host' is used to identify the Windows process that called this script
#
# KEY POINT:
#  This script must be executed by Windows Task Scheduler AT THE SAME FREQUENCY 
#  as the WATCHDOG APP USED TO CHECK FOR RAM LEAK EVENTS.
#
# Version: 1
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V1 - 2025-02-22 - initial release
#
# USAGE:      {this script name}.ps1 [arg1]
# ARGUMENTS:  -Current       = [OPTIONAL] show current memory usage parameters (no logging)
#             -SimLeak       = [OPTIONAL] like '5X', '400MB' OR '5X', '10%' OR '5X', '400MB', '10%'
#             -RemSimEntries = [OPTIONAL] remove entries created by argument -SimLeak from the CSV log file
#             -Help          = display usage guide
#             -Examples      = display examples usage guide
# RETURNS:    [String] 'success' or err msg
#
# BEFORE USING THIS SCRIPT:
#   1. EDIT ALL VARIABLES IN THE 'USER SETTINGS' FILE
#   2. AUTOMATE THIS SCRIPT VIA THE METHOD BELOW
#
# HOW TO AUTOMATE THIS SCRIPT VIA TASK SCHEDULER
#   Open Task Scheduler and create an appropriately named task with one trigger and one action:
#   Task Scheduler
#   'Triggers' tab:
#       NAVIGATE TO 'Begin the task' > SELECT 'On a schedule' 
#       NAVIGATE TO 'Advanced settings'
#       CHECK BOX 'Repeat task every' > SELECT (OR TYPE) 'X minutes' > 'for a duration of' SELECT 'indefinitely'
#       (I'm currently using X=3)
#   'Actions' tab:
#       Program/script: powershell.exe
#       Add arguments: {your_powershell_scripts_path)\{script_file_name}.ps1

# MUST BE THE FIRST LINE IN A SCRIPT
param(
	[Switch]$Current,        # [OPTIONAL] show current memory usage parameters (no logging)
	[String[]]$SimLeak,      # [OPTIONAL] like '5X', '400MB' OR '5X', '10%' OR '5X', '400MB', '10%'
	[Switch]$RemSimEntries,  # [OPTIONAL] remove entries created by argument -SimLeak from the CSV log file
	[Switch]$Help,           # [OPTIONAL] show usage guide
	[Switch]$Examples        # [OPTIONAL] display script usage examples
)

Set-StrictMode  -Version 3  # https://tinyurl.com/mrx3w9v4
#Start-Transcript  # default destination is C:\Users\{user}l\Documents

$ScriptVersion = "V1"

# FUNCTIONS
if (1 -eq 1) {  # Use to collapse this code section while editing

	function UpdateFileStack() {
	# PURPOSE:  Update stack of log files with names like: 'bi_ram_log[1].csv' .. 'bi_ram_log[N].csv'
	# ARGS:     see param lines below
	# RETURNS:  (string) actual stack count or 'ERR/...' message
		
		param (
			[String]$SourcePath,
			[String]$FileNameBase,
			[String]$Needle,
			[Int]$StackCnt,
			[String]$DestPath,          # [OPTIONAL] if blank, stack is made in $SourcePath
			[Switch]$MakeStackInSource  # [OPTIONAL] $true makes stack in *both* $SourcePath and $DestPath
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "SourcePath = '$SourcePath'"
		# Write-Host "FileNameBase = '$FileNameBase'"
		# Write-Host "Needle = '$Needle'"
		# Write-Host "StackCnt = StackCnt"
		
		if (!($SourcePath -and $FileNameBase -and $Needle -and $StackCnt)) {
			$err_msg = "ERR/missging required arguments: -SourcePath -FileNameBase -Needle or -StackCnt"
			return $err_msg
		}
		
		if (!$DestPath) {
			$make_stack_in_source_flg = $true
		} else {
			$make_stack_in_source_flg = $MakeStackInSource
		}

		# DELETE EXISTING STACK
		# ... Why? in case user later reduces $maxStackFiles for an existing stack.
		# ... Note: take care when filtering w/ square brackets, see https://stackoverflow.com/a/21009701/13499984
		Get-ChildItem $DestPath | Where-Object { $_.Name -like "$FileNameBase``[*" } | Remove-Item -force
		if ($make_stack_in_source_flg) {
			Get-ChildItem $SourcePath | Where-Object { $_.Name -like "$FileNameBase``[*" } | Remove-Item -force			
		}
		
		# SORT FILES IN DESCENDING ORDER AND USE ONLY THE TOP 'N'
		$j = 1  # Initialize stack index
		$k = 1  # Initialize stack index
		
		# ... Rather than sort by the file's last-write date 
		#     i.e., 'Sort LastWriteTime -Descending)'
		# ... we sort by the 'date-like' substrings in the filename.
		#     WHY? On occasion, the user may need to manually edit a CSV file; 
		#          (it is not rare that a CSV log file contains erroneous entries)
		# ... HEADS-UP - this line of code can be CPU-intensive (see Notes above)
		@(Get-ChildItem $SourcePath -Filter $FileNameBase$Needle.csv | Sort -Descending) | select -First $StackCnt | % {

			# CREATE STACK ENTRY
			# ... in the destination folder
			Copy-Item -path $_.FullName -destination $("$DestPath\$FileNameBase[$j].csv") -force
			$j++
			# ... in the source folder
			if ($make_stack_in_source_flg) {
				Copy-Item -path $_.FullName -destination $("$SourcePath\$FileNameBase[$k].csv") -force
				$k++
			}
		}
		
		# ACTUAL STACK COUNT
		$return_val = $j-1
		
		return $return_val.ToString()
		
	}
	
	function UpdateCsvFile() {
	# PURPOSE:  Log latest RAM usage data to CSV file.
	# ARGS:     see param lines below
	# RETURNS:  (string) 'success' or 'ERR/...' message
	# NOTES:    1. CSV file is created automatically if it does not already exist. 
	#           2. CSV file's header row is generated from the object keys in variable $NewRowObj.
		
		param (
			[String]$CsvPath,
			[PSCustomObject]$NewRowObj
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		
		# APPEND NEW ROW TO CSV FILE, creating the file if necessary
		if ($PSVersionTable.PSVersion.Major -ge 7) {
			# REFS: https://tinyurl.com/yvy4mjy6
			#       https://stackoverflow.com/a/70717331
			#       https://stackoverflow.com/a/60681762
			try {
				# NOTE: To simulate an Export-Csv error, remove the -Force argument and make file read-only
				$NewRowObj | Export-Csv -LiteralPath $CsvPath -Append -Force -UseQuotes AsNeeded
			} catch {
				$err_msg = "ERR/unable to append row to CSV file; $_"
				return $err_msg
			}
		} else {
			try {
				# NOTE: To simulate an Export-Csv error, remove the -Force argument and make file read-only
				$NewRowObj | Export-Csv -LiteralPath $CsvPath -Append -Force -NoTypeInformation
			} catch {
				$err_msg = "ERR/unable to append row to CSV file; $_"
				return $err_msg
			}
		}

		if (!(Test-Path -LiteralPath $CsvPath)) {
			$err_msg = "ERR/unable to write to CSV log ($CsvPath)"
			return $err_msg
		}
		
		return 'success'
		
	}

	function Format-Color() {
	# PURPOSE:  Colorize the command's output based on line-by-line pattern matching.
	#           Also supports pausing output based on console's line height
	# SOURCE:   https://www.bgreco.net/powershell/format-color/
	# REFS:     https://superuser.com/a/1709091  (colorize output)
	#           https://stackoverflow.com/a/4647985/13499984  (colorize output)
	#           https://stackoverflow.com/a/44872079/13499984  (measure window height)
	# REQUIRES: PauseOutput()

		param (
			[Hashtable]$InputObject = @{}
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		$Colors      = $InputObject.Colors
		$SimLeakpleMatch = $InputObject.SimpleMatch
		$PauseOutput = $InputObject.PauseOutput

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host ("$fn_name() -InputObject =" + ($InputObject | ConvertTo-Json))
		# Write-Host ("Colors type = " + $Colors.GetType().Name)
		# Write-Host ("Colors = " + ($Colors | ConvertTo-Json))
		# Write-Host "SimpleMatch = $SimLeakpleMatch"
		# Write-Host "PauseOutput = $PauseOutput"
		# Write-Host ("console_lines = " + $Host.UI.RawUI.WindowSize.Height)

		$output_lines = ($input | Out-String) -replace "`r", "" -split "`n"
		$i = 0
		foreach($line in $output_lines) {
			$color = ''
			foreach($pattern in $Colors.Keys){
				if(!$SimLeakpleMatch -and $line -match $pattern) { 
					$color = $Colors[$pattern] 
				} elseif ($SimLeakpleMatch -and $line.IndexOf($pattern) -gt -1) { 
					$color = $Colors[$pattern]
				}
			}

			if($color) {
				Write-Host $line -ForegroundColor $color 			
			} else {
				Write-Host $line
			}

			if ($PauseOutput) {
				if ($Host.UI.SupportsVirtualTerminal) {
					# KEY-PONT: This feature supports only Consoles.
					# The Powershell ISE UI does not have a height because it is a Forms window. Its viewport 
					# has a size, but it is not tied to the number of visible lines
					$console_lines = $Host.UI.RawUI.WindowSize.Height
					if ($i -eq $console_lines - 5) {
						PauseOutput('Press any key to continue...')
					}
				}
			}

			$i++
		}
	}

}

# IMPORT USER SETTINGS...
# Note: The 'user settings' script is expected to be in same folder as this script
$ScriptName = $MyInvocation.MyCommand.Name
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	return "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
}

# ------------------------------------
# VALIDATE USER SETTINGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	$variable_test = Get-Variable -Name csv_logfile_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$csv_logfile_path"
		return $err_msg
	} else {
		if ($csv_logfile_path.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$csv_logfile_path is not set"
			return $err_msg
		}
	}
	$csv_logfile_path = $csv_logfile_path.Replace("/","\").TrimEnd('\')

	$variable_test = Get-Variable -Name csv_logfile_basename -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$csv_logfile_basename = 'bi_ram_log'
	}

	$variable_test = Get-Variable -Name update_minutes -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$update_minutes = 3
	}

}
# ------------------------------------

# GATHER SCRIPT & ENVIRONMENT INFO 
if (1 -eq 1) {
	
	#$ScriptName = $MyInvocation.MyCommand.Name
	#$ScriptName = (Get-Variable MyInvocation).Value.MyCommand.Name
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')

	# For debugging
	# Write-Host "ScriptName = $ScriptName"
	# Write-Host "ScriptPath = $ScriptPath"
	# Write-Host "ScriptBaseName = $ScriptBaseName"
	# Write-Host "ScriptPath = $ScriptPath"

	# INITIALIZE VARS
	$PsVersion = $PSVersionTable.PSVersion.Major

	# For testing
	# Write-Host "ScriptName = $ScriptName"
	# Write-Host "PsVersion = $PsVersion"

	# NOTE: Scripts running from the following hosts have process names as follows:
	#       Task Scheduler is 'svchost', Powershell-ISE is 'explorer', Windows Terminal is 'WindowsTerminal'
	#       https://stackoverflow.com/a/65894924
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	# ... create lookup table (used by script's CSV log)
	$HostLabels = @{}
	$HostLabels['svchost'] = 'TaskScheduler'
	$HostLabels['explorer'] = 'Powershell or Powershell-ISE'
	$HostLabels['WindowsTerminal'] = 'WindowsTerminal'

	# For testing
	# Write-Host "ParentProcess = $ParentProcess"

	# ... Get friendly host name
	if ($HostLabels[$ParentProcess]) {
		$script_host_str = $HostLabels[$ParentProcess]
	} else {
		$script_host_str = $ParentProcess
	}

	# For testing
	# Write-Host "ParentProcess = $ParentProcess"

	# NOTE: Scripts called by another script have a 'PSCommandPath'.
	#       Method 'PSCommandPath' returns null if script is executed via the console
	#       https://stackoverflow.com/a/23370618/13499984
	$CallingScript = $MyInvocation.PSCommandPath

	# For testing
	# Write-Host "CallingScript = $CallingScript"

	# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM BLUE IRIS)
	$FromConsole = !(($ParentProcess -eq 'svchost' -or $ParentProcess -eq 'BlueIris'))

	# For testing
	# $FromConsole = $false

}

# SHOW HELP
if ($PSBoundParameters.ContainsKey('Help') -and $FromConsole) {
	Write-Host "================================================================================================"
	Write-Host "BLUE IRIS RAM LOGGER UTILITY"
	Write-Host "- Creates and maintains a CSV log file that has these columns:"
	Write-Host "  Timestamp, SoftwareUsedRamBytes, Uptime, RamPct, Tag, Host"
	Write-Host "------------------------------------------------------------------------------------------------"
	Write-Host "Usage: $ScriptName [No args] ... adds new row to the CSV file [1,2]"
	Write-Host "       $ScriptName -Examples ... see more options"
	Write-Host "Optional Args:"
	Write-Host "   -Current ... show current memory usage parameters (no logging)"
	Write-Host "   -SimLeak ... simulate a memory leak by adding new entries with stepwise increases [3]"
	Write-Host "   -RemSimEntries ... remove entries created by argument -SimLeak from the CSV log file"
	Write-Host "   -Help     ... show this usage guide"
	Write-Host "   -Examples ... show the examples guide"
	Write-Host "Notes: "
	Write-Host ("[1] The CSV log file is created monthly w/ a filename like '" + $csv_logfile_basename + "_202411.csv';")
	#Write-Host "    The file has these column headers:"
	#Write-Host "    'Timestamp','SoftwareUsedRamBytes','Uptime','RamPct','Tag','Host'"
	Write-Host "[2] Column 'SoftwareUsedRamBytes' is the memory used by the active Blue Iris process"
	Write-Host "    Column 'Uptime' is the uptime for the active Blue Iris process"
	Write-Host "    Column 'RamPct' is the memory used by the system"
	Write-Host "    Column 'Tag' differentiates between automated- and console-generated entries"
	Write-Host "    Column 'Host' differentiates the script's environment (Task Scheduler vs the console)"
	Write-Host "[3] Adds simulated memory leak entries to the CSV file (for testing watchdog utilities)"
	Write-Host "    like, -SimLeak '5X', '400MB' OR -SimLeak '5X', '10%'"
	Write-Host "    * term 'NX' appends N new 'simulation' tagged entries to the CSV file"
	Write-Host "    * each new 'Timestamp' value is incremented by `$update_minutes (via the user-setting-file)"
	Write-Host "    * each new 'SoftwareUsedRamBytes' value is incremented by the specified number of MB [4]"
	Write-Host "    * each new 'RamPct' value is incremented by the specified % (up to 100%)"
	Write-Host "    * quotes around the argument's values are not requested"
	Write-Host "[4] Up to the installed RAM of the machine"
	Write-Host "------------------------------------------------------------------------------------------------"
	Write-Host ""
	exit
}

# SHOW EXAMPLES
if ($PSBoundParameters.ContainsKey('Examples') -and $FromConsole) {
	$examples_str = `
	"=================================================================================================" `
	+ "`nBLUE IRIS RAM LOGGER UTILITY - USAGE EXAMPLES" `
	+ "`n--> to try these commands, copy/paste the colored text to the console prompt" `
	+ "`n-------------------------------------------------------------------------------------------------" `
	+ "`n  1. Show current memory usage parameters (no logging to CSV file)" `
	+ "`n     $ScriptName -Current"`
	+ "`n  2. Update the CSV log file with 1 new entry" `
	+ "`n     $ScriptName"`
	+ "`n  3. Append 5 new rows simulating a stepwise 400MB increase in the software used RAM" `
	+ "`n     $ScriptName -SimLeak 5X, 400MB"`
	+ "`n  4. Append 5 new rows simulating a stepwise 10% increase in the system used RAM%" `
	+ "`n     $ScriptName -SimLeak 5X, 10%"`
	+ "`n  5. Like #3 & #4 combined" `
	+ "`n     $ScriptName -SimLeak 5X, 400MB, 10%"`
	+ "`n  6. Remove all 'simulation'-tagged entries from the CSV file" `
	+ "`n     $ScriptName -RemSimEntries" `
	+ "`n-------------------------------------------------------------------------------------------------"

	# For inline hashtables, use unquoted keys and a semicolon separator
	# https://learn.microsoft.com/en-us/powershell/scripting/learn/deep-dives/everything-about-hashtable?view=powershell-7.4
	$InputObject = @{
		Colors = @{ "$ScriptName" = 'Cyan' };
		SimpleMatch = $true;
		PauseOutput = $true
	}
	# We use an object in the following pipe because 
	# REF: https://stackoverflow.com/a/53780647/13499984
	$examples_str | Format-Color $InputObject
	
	
	exit
}

# HANDLE ADD SIMULATION ENTRIES REQUEST
$sim_mode_flg = $false
if ($PSBoundParameters.ContainsKey('SimLeak')) {

	# Make string of params (used for logging to the script log & the Blue Iris logfile)

	# For debugging
	# Write-Host ("SimLeak = " + ($SimLeak | ConvertTo-Json))
	# Write-Host ("SimLeak type = " + $SimLeak.GetType())
	# Write-Host ("SimLeak[1] type = " + $SimLeak[1].GetType())
	# Write-Host ("SimLeak[2] type = " + $SimLeak[2].GetType())

	if ($SimLeak.Count -ge 2 -or $SimLeak.Count -le 3) {
		
		$sim_mode_flg = $true
		$sim_N = [Int]($SimLeak[0] -replace "[^0-9]", "")
		$sim_mb = 0
		$sim_pct = 0
		if ($SimLeak[1].IndexOf('%') -ge 0) {
			if ($SimLeak[1].IndexOf('.') -lt 0) {
				$sim_pct = [Int]($SimLeak[1] -replace "[^0-9]", "")
			} else {
				$err_msg = "ERR/percent value must an integer; fractions not allowed"
				return $err_msg
			}
		} else {
			if ($SimLeak[1].IndexOf('.') -lt 0) {
				$sim_mb = [Int]$SimLeak[1]
			} else {
				$err_msg = "ERR/MB value must an integer; fractions not allowed"
				return $err_msg
			}
		}
		if ($SimLeak.Count -eq 3) {
			if ($SimLeak[2].IndexOf('%') -ge 0) {
				if ($SimLeak[2].IndexOf('.') -lt 0) {
					$sim_pct = [Int]($SimLeak[2] -replace "[^0-9]", "")
				} else {
					$err_msg = "ERR/percent value must an integer; fractions not allowed"
					return $err_msg
				}
			} else {
				if ($SimLeak[2].IndexOf('.') -lt 0) {
					$sim_mb = [Int64]$SimLeak[2]
				} else {
					$err_msg = "ERR/MB value must an integer; fractions not allowed"
					return $err_msg
				}
			}
		}

		if ($sim_mb -eq 0 -and $sim_pct -eq 0) {
			$err_msg = "ERR/expecting argument -SimLeak to be like '5X', '400MB' OR '5X', '10%' OR '5X', '400MB', '10%'"
			return $err_msg			
		}

	} else {
		$err_msg = "ERR/expecting argument -SimLeak to be like '5X', '400MB' OR '5X', '10%' OR '5X', '400MB', '10%'"
		return $err_msg			
	}
	
	# For debugging
	# Write-Host "sim_mode_flg = $sim_mode_flg"
	# Write-Host "sim_N = $sim_N"
	# Write-Host ("sim_mb = $sim_mb (type = " + $sim_mb.GetType() + ")")
	# Write-Host ("sim_pct = $sim_pct (type = " + $sim_pct.GetType()) + ")"

}

# INITIALIZE
$process_name = "BlueIris"
$YYYYMM = Get-Date -Format "yyyyMM"
$csvfilespec = $csv_logfile_path + "\" + $csv_logfile_basename + "_" + $YYYYMM + ".csv"
$tag_str = 'periodic'  # Default
if (@('explorer', 'WindowsTerminal') -contains $ParentProcess) {
	$tag_str = 'console'
}

# REMOVE SIMULATION ENTRIES FROM CSV FILE
if ($RemSimEntries) {
	
	# Import the CSV file
	$csv_rows = Import-Csv $csvfilespec
	
	# For debugging'
	# Write-Host ("csv_rows type = " + $csv_rows.Gettype())

	# Filter out rows where the 4th column contains the string
	$csv_rows_new = $csv_rows | Where-Object { $_.Tag -ne "simulation" } 

	# Export the filtered data back to the CSV file
	$csv_rows_new | Export-Csv $csvfilespec -NoTypeInformation
	
	$rows_removed_cnt = $csv_rows.Count - $csv_rows_new.count
	
	if ($rows_removed_cnt -gt 0) {
		if ($FromConsole) {
			Write-Host "CSV file purged of all 'simulation' tagged entries ($rows_removed_cnt total)"
		}
	} else {
		return "Nothing to do - no 'simulation' tagged entries found in the CSV file"
	} 

	# UPDATE CSV LOG FILE STACK
	if (1 -eq 1) {  # Use to collapse this code section while editing
		if ($rows_removed_cnt -gt 0) {
			$dest_path = ''
			$make_stack_in_source_flg = $true

			$result = UpdateFileStack `
				-SourcePath $csv_logfile_path `
				-FileName $csv_logfile_basename `
				-Needle "*_20*" `
				-DestPath "" `
				-StackCnt 6
				
			if (!$result.StartsWith('ERR/')) {
				$x = [Int]$result
				if ($FromConsole) {
					if ($dest_path) {
						if (Test-Path -LiteralPath $dest_path) {
							#Write-Host "Created/updated $x '$csv_logfile_basename[i].csv' stack files in BOTH source & destination folders"
							Write-Host "Updated the latest stack files in BOTH source & destination folders"
						}
					} else {
						#Write-Host "Created/updated $x '$csv_logfile_basename[i].csv' stack files in source folder"
						Write-Host "Updated the latest stack file in the source folder"		
					}
					Write-Host ""
				}
				exit
			} else {
				return "ERR/Unable to create logfile stack: '$result'"
			}
		}
	}
		
	if (!$FromConsole) {
		return 'success'
	}
}

if ($FromConsole) {
	Write-Host ""
	Write-Host "Current Memory Usage Parameters:"
	Write-Host "--------------------------------"
}

# GET SYSTEM-USED-RAM-TOTAL
if (1 -eq 1) {  # Use to collapse this code section while editing
	#$os_obj = Get-CimInstance -ClassName Win32_OperatingSystem
	#$used_ram_kb = ($os_obj.TotalVisibleMemorySize - $os_obj.FreePhysicalMemory)

	# Write-Host ("System Used RAM = {0:N0} bytes" -f $used_ram)
	# Write-Host ("System Used RAM = " + (Format-BytesToMBGB $used_ram))
}

# GET SYSTEM RAM AND SYSTEM-USED-RAM-PERCENT
if (1 -eq 1) {  # Use to collapse this code section while editing

	$cs_obj = Get-CimInstance -ClassName Win32_ComputerSystem
	$total_ram_bytes = $cs_obj.TotalPhysicalMemory  # Installed RAM in bytes
	$total_ram_kb = ($total_ram_bytes / 1KB)
	$total_ram_mb = ($total_ram_bytes / 1MB)
	$total_ram_gb = ($total_ram_bytes / 1GB)

	$os_obj = Get-CimInstance -ClassName Win32_OperatingSystem
	$total_ram_kb = $os_obj.TotalVisibleMemorySize  # OS-visible RAM in KB
	$free_ram_kb = $os_obj.FreePhysicalMemory       # in KB
	$used_ram_kb = $total_ram_kb - $free_ram_kb # in KB
	
	# For Debugging
	#Write-Host "system total RAM = $total_ram_gb GB"

	# Calculate the percentage of used RAM
	$pct_used = ($used_ram_kb / $total_ram_kb) * 100
	$pct_used_str = "{0:N0}%" -f $pct_used

	if ($FromConsole) {
		$str_bytes = $total_ram_bytes.ToString("N0")
		$str_mb    = ($total_ram_bytes / 1MB).ToString("N2")
		$str_gb    = ($total_ram_bytes / 1GB).ToString("N3")
		Write-Host ("Physical RAM installed (system):        $str_bytes bytes ($str_mb MB) ($str_gb GB)")
		Write-Host ("Physical RAM in use (system):           $pct_used_str")
	}
}

# GET BLUE IRIS PROCESS-USED-RAM
if (1 -eq 1) {  # Use to collapse this code section while editing

	# KEY POINT: We use 64-bit signed integer properties because system memory can easily
	# exceed the maximum value for a 32-bit signed integer (2,147,483,647 bytes).
	# Using 32-bit signed integer properties can result in NEGATIVE values due to a value  
	# that "wraps around" into the negative range due to integer overflow. 
	# For example:
	# WorkingSet = 2,147,483,648 bytes (2 GiB) ==> -2,147,483,648 (overflowed to negative).

	$process_obj = Get-Process -Name $process_name -ErrorAction SilentlyContinue

	# For debugging
	# Get-Process -Name $process_name
	# Write-Host ("process_obj = " + ($process_obj | ConvertTo-Json))

	# In bytes
	$all_processes_ram_obj = $process_obj | 
		Measure-Object -Property WorkingSet -Sum

	# Looks like...
	# {
	#	"Count": 2,
	#	"Average": null,
	#	"Sum": 1267793920.0,
	#	"Maximum": null,
	#	"Minimum": null,
	#	"StandardDeviation": null,
	#	"Property": "WorkingSet"
	# }	

	# For debugging
	# Write-Host ("all_processes_ram_obj type = " + $all_processes_ram_obj.GetType())
	# Write-Host ("all_processes_ram_obj = " + ($all_processes_ram_obj | ConvertTo-Json))
	# Write-Host ("all_processes_ram_obj.Sum type = " + $all_processes_ram_obj.Sum.GetType())
	# Write-Host ("all_processes_ram_obj.Sum = " + $all_processes_ram_obj.Sum)

	# In bytes
	$largest_process_ram_obj = $process_obj | 
		Sort-Object -Property WorkingSet64 -Descending |
		Select-Object -First 1

	# For debugging
	# Write-Host ("largest_process_ram_obj type = " + $largest_process_ram_obj.GetType())
	# Write-Host ("largest_process_ram_obj = " + ($largest_process_ram_obj | ConvertTo-Json))
	# Write-Host ("largest_process_ram_obj.WorkingSet64 type = " + $largest_process_ram_obj.WorkingSet64.GetType())
	# Write-Host ("largest_process_ram_obj.WorkingSet64 = " + ($largest_process_ram_obj.WorkingSet64 | ConvertTo-Json))
	# Write-Host ("largest_process_ram_obj.WorkingSet64 type = " + $largest_process_ram_obj.WorkingSet64.GetType())
	# Write-Host ("largest_process_ram_obj.WorkingSet64 = " + ($largest_process_ram_obj.WorkingSet64 | ConvertTo-Json))

	if ($FromConsole) {
		if ($null -ne $all_processes_ram_obj.Sum) {
			$str_bytes = $all_processes_ram_obj.Sum.ToString("N0")
			$str_mb    = ($all_processes_ram_obj.Sum / 1MB).ToString("N2")
			$str_gb    = ($all_processes_ram_obj.Sum / 1GB).ToString("N3")
			Write-Host ("RAM used by all ${process_name} processes:     $str_bytes bytes ($str_mb MB) ($str_gb GB)")

			if ($null -ne $largest_process_ram_obj.WorkingSet64) {
				$str_bytes = $largest_process_ram_obj.WorkingSet64.ToString("N0")
				$str_mb    = ($largest_process_ram_obj.WorkingSet64 / 1MB).ToString("N2")
				$str_gb    = ($largest_process_ram_obj.WorkingSet64 / 1GB).ToString("N3")
				Write-Host ("RAM used by main ${process_name} process:      $str_bytes bytes ($str_mb MB) ($str_gb GB)")
				# Write-Host ("RAM used by main ${process_name} process = {0:N0} bytes" -f ($largest_process_ram_obj.WorkingSet64))
				# Write-Host ("RAM used by main ${process_name} process = {0:N2} MB" -f ($largest_process_ram_obj.WorkingSet64 / 1MB))
				# Write-Host ("RAM used by main ${process_name} process = {0:N3} GB" -f ($largest_process_ram_obj.WorkingSet64 / 1GB))
			}
			
		} else {
			Write-Host "Process $process_name not found or not consuming memory."
		}
	}
}

# GET PROCESS-UPTIME
$uptime_str = ""
if ($process_obj.StartTime) {

	# KEY-POINT:
	#    $process_obj.StartTime will return TWO values if both the Blue Iris service and application are running.
	#    We use the minimum StartTime, assuming the BI service is loaded first
	[Array]$start_times = $process_obj.StartTime
	if ($start_times.Count -gt 1) {
		$datetime_min = [DateTime]::MaxValue  # returns 12/31/9999 23:59:59
		foreach ($datetime in $start_times) {
			$datetime_obj = [DateTime]::Parse($datetime)
			if ($datetime_obj -lt $datetime_min) {
				$datetime_min = $datetime_obj
			}
		}
		$start_time = $datetime_min
	} else {
		$start_time = $start_times[0]
	}
	$uptime_obj = (Get-Date) - $start_time

	# For debugging
	# Write-Host ("process_obj.StartTime = " + ($process_obj.StartTime | ConvertTo-Json))
	# Write-Host ("start_time = " + $start_time)

	$uptime_str = "{0}:{1:D2}:{2:D2}:{3:D2}" -f $uptime_obj.Days, $uptime_obj.Hours, $uptime_obj.Minutes, $uptime_obj.Seconds
	
	# Detect if server recently restarted
	if ($uptime_obj.TotalMinutes -le $update_minutes) {

		$tag_str = 'restart'
		
		if ($FromConsole) {
			Write-Host "Process $process_name was recently restarted"
		}
	}

	if ($FromConsole) {
		Write-Host ("Uptime for 1st-loaded $process_name process: $uptime_str (total minutes: " + $uptime_obj.TotalMinutes.ToString("0") + ")")
	}

	# For debugging
	# Write-Host ("process_obj.StartTime = " + $process_obj.StartTime)
	# Write-Host ("uptime type = " + $uptime_obj.GetType())
	# Write-Host ("uptime.Minutes type = " + $uptime_obj.Minutes.GetType())
	# Write-Host ("uptime.TotalMinutes = " + $uptime_obj.TotalMinutes)
	# Write-Host ("update_minutes = " + $update_minutes)
	# Write-Host "uptime_obj for $process_name process = $uptime_obj"
	
	
} else {
	if ($FromConsole) {
		Write-Host "Process $process_name not found."
	}
}


if ($Current) {
	if ($FromConsole) {
		Write-Host ""
	} else {
		return "success"
	}
}

if ($FromConsole) {
	Write-Host ""
	Write-Host "Log File Update:"
	Write-Host "----------------"
}

# UPDATE THE CSV LOGFILE
if (1 -eq 1) {  # Use to collapse this code section while editing

	if ($FromConsole) {
		Write-Host "New row appended to: '$csvfilespec'"
	}

	if (!$sim_mode_flg) {  # Add standard entries

		$timestamp = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
			
		# Note: this object's keys are the header column labels
		$new_row_obj = [PSCustomObject]@{
			'Timestamp'            = $timestamp
			'SoftwareUsedRamBytes' = $largest_process_ram_obj.WorkingSet64
			'Uptime'               = $uptime_str
			'RamPct'               = $pct_used_str
			'Tag'                  = $tag_str
			'Host'                 = $script_host_str
		}
		
		if ($FromConsole) {
			Write-Host ("new row = " + ($new_row_obj | ConvertTo-Json))
		}
				
		$response_str = `
			UpdateCsvFile `
			-CsvPath $csvfilespec `
			-NewRowObj $new_row_obj
			
		if ($response_str.StartsWith('ERR/')) {
			Write-Warning "unable to update CSV file: $response_str"
		}

	} 
	else {                 # Add simulation entries

		$now = Get-Date	
		
		for ($i=0; $i -lt $sim_N; $i++) {

			if ($sim_mb -or $sim_pct) {

				$timestamp = ($now.AddMinutes($i * $update_minutes))
				$timestamp = $timestamp.ToString("yyyy/MM/dd_HH:mm:ss")

				if ($sim_mb) {
					$sw_used_bytes = [Math]::Min($largest_process_ram_obj.WorkingSet64 + $i * $sim_mb, $total_ram_bytes)
				} else {
					$sw_used_bytes = $largest_process_ram_obj.WorkingSet64
				}
				if ($sim_pct) {
					$pct_used = ($used_ram_kb / $total_ram_kb) * 100
					$pct_used_adj = [Math]::Min($pct_used + $i * $sim_pct, 100)
					$pct_used_str = "{0:N0}%" -f $pct_used_adj
				}

				# Note: this object's keys are the header column labels
				$new_row_obj = [PSCustomObject]@{
					'Timestamp'            = $timestamp
					'SoftwareUsedRamBytes' = $sw_used_bytes.ToString()
					'Uptime'               = $uptime_str
					'RamPct'               = $pct_used_str
					'Tag'                  = 'simulation'
					'Host'                 = $script_host_str
				}
		
				if ($FromConsole) {
					Write-Host ("CSV file new row = " + ($new_row_obj | ConvertTo-Json))
				}
				
				$response_str = `
					UpdateCsvFile `
					-CsvPath $csvfilespec `
					-NewRowObj $new_row_obj
					
				if ($response_str.StartsWith('ERR/')) {
					Write-Warning "unable to update CSV file: $response_str"
				}
			}
		}
	}

	if (!(Test-Path -LiteralPath $csvfilespec)) {
		$err_msg = "ERR/unable to write to CSV log ($csvfilespec)"
		return $err_msg
	}
}

if ($FromConsole) {
	Write-Host ""
	Write-Host "Log File 'Stack':"
	Write-Host "-----------------"
}

# UPDATE CSV LOGFILE 'STACK'
if (1 -eq 1) {  # Use to collapse this code section while editing

	$dest_path = ''
	$make_stack_in_source_flg = $true

	$result = UpdateFileStack `
		-SourcePath $csv_logfile_path `
		-FileName $csv_logfile_basename `
		-Needle "*_20*" `
		-DestPath $dest_path `
		-StackCnt 6
		
	if (!$result.StartsWith('ERR/')) {
		$x = [Int]$result
		if ($FromConsole) {
			Write-Host "Created/refreshed $x '$csv_logfile_basename[i].csv' stack files in folder '$csv_logfile_path'"		
		}
	} else {
		Write-Host "Unable to create logfile stack: '$result'"		
	}
}

if ($FromConsole) {
	Write-Host ""
} else {
	return "success"
}
